@@ -47,7 +47,6 @@ fbPolyArc (DrawablePtr	pDrawable,
     
     if (pGC->lineWidth == 0)
     {
-#ifndef FBNOPIXADDR
 	arc = 0;
 	if (pGC->lineStyle == LineSolid && pGC->fillStyle == FillSolid)
 	{
@@ -55,9 +54,7 @@ fbPolyArc (DrawablePtr	pDrawable,
 	    {
 	    case 8:	arc = fbArc8; break;
 	    case 16:    arc = fbArc16; break;
-#ifdef FB_24BIT
 	    case 24:	arc = fbArc24; break;
-#endif
 	    case 32:    arc = fbArc32; break;
 	    }
 	}
@@ -71,15 +68,12 @@ fbPolyArc (DrawablePtr	pDrawable,
 	    BoxRec	box;
 	    int		x2, y2;
 	    RegionPtr	cclip;
-	    int		wrapped = 0;
+#ifdef FB_ACCESS_WRAPPER
+	    int		wrapped = 1;
+#endif
 	    
 	    cclip = fbGetCompositeClip (pGC);
 	    fbGetDrawable (pDrawable, dst, dstStride, dstBpp, dstXoff, dstYoff);
-#ifdef FB_ACCESS_WRAPPER
-	    wrapped = 1;
-#else
-	    wrapped = 0;
-#endif
 	    while (narcs--)
 	    {
 		if (miCanZeroArc (parcs))
@@ -141,7 +135,6 @@ fbPolyArc (DrawablePtr	pDrawable,
 #endif
 	}
 	else
-#endif
 	    miZeroPolyArc (pDrawable, pGC, narcs, parcs);
     }
     else
